@@ -3,44 +3,49 @@
 //! Under the hood, this makes use of Git's advanced rebase functionality, which
 //! is also used to preserve merge commits using the `--rebase-merges` option.
 
+use std::convert::TryFrom;
 use std::fmt::Write;
 use std::time::SystemTime;
 
+use eden_dag::DagAlgorithm;
 use tracing::instrument;
 
 use crate::core::config::get_restack_preserve_timestamps;
+use crate::core::dag::{resolve_commits, CommitSet, Dag, ResolveCommitsResult};
+use crate::core::effects::Effects;
 use crate::core::eventlog::{EventLogDb, EventReplayer};
-use crate::core::graph::{
-    make_graph, resolve_commits, BranchOids, CommitGraph, HeadOid, MainBranchOid,
-    ResolveCommitsResult,
-};
-use crate::core::mergebase::MergeBaseDb;
 use crate::core::rewrite::{
-    execute_rebase_plan, BuildRebasePlanOptions, ExecuteRebasePlanOptions, RebasePlanBuilder,
+    execute_rebase_plan, BuildRebasePlanOptions, ExecuteRebasePlanOptions, ExecuteRebasePlanResult,
+    RebasePlanBuilder,
 };
 use crate::git::{GitRunInfo, NonZeroOid, Repo};
-use crate::tui::Effects;
+use crate::opts::MoveOptions;
 
 #[instrument]
 fn resolve_base_commit(
-    graph: &CommitGraph,
+    dag: &Dag,
     merge_base_oid: Option<NonZeroOid>,
     oid: NonZeroOid,
-) -> NonZeroOid {
-    let node = &graph[&oid];
-    if node.is_main {
-        oid
-    } else {
-        match node.parent {
-            Some(parent_oid) => {
-                if graph[&parent_oid].is_main || Some(parent_oid) == merge_base_oid {
-                    oid
-                } else {
-                    resolve_base_commit(graph, merge_base_oid, parent_oid)
-                }
-            }
-            None => oid,
+) -> eyre::Result<NonZeroOid> {
+    let bases = match merge_base_oid {
+        Some(merge_base_oid) => {
+            let range = dag
+                .query()
+                .range(CommitSet::from(merge_base_oid), CommitSet::from(oid))?;
+            let roots = dag.query().roots(range.clone())?;
+            let bases = dag.query().children(roots)?.intersection(&range);
+            bases
         }
+        None => {
+            let ancestors = dag.query().ancestors(CommitSet::from(oid))?;
+            let bases = dag.query().roots(ancestors)?;
+            bases
+        }
+    };
+
+    match bases.first()? {
+        Some(base) => NonZeroOid::try_from(base),
+        None => Ok(oid),
     }
 }
 
@@ -52,10 +57,7 @@ pub fn r#move(
     source: Option<String>,
     dest: Option<String>,
     base: Option<String>,
-    force_in_memory: bool,
-    force_on_disk: bool,
-    dump_rebase_constraints: bool,
-    dump_rebase_plan: bool,
+    move_options: &MoveOptions,
 ) -> eyre::Result<isize> {
     let repo = Repo::from_current_dir()?;
     let head_oid = repo.get_head_info()?.oid;
@@ -90,67 +92,64 @@ pub fn r#move(
             }
         },
     };
-    let (source_oid, dest_oid) = match resolve_commits(&repo, vec![source, dest])? {
-        ResolveCommitsResult::Ok { commits } => match &commits.as_slice() {
-            [source_commit, dest_commit] => (source_commit.get_oid(), dest_commit.get_oid()),
-            _ => eyre::bail!("Unexpected number of returns values from resolve_commits"),
-        },
-        ResolveCommitsResult::CommitNotFound { commit } => {
-            writeln!(effects.get_output_stream(), "Commit not found: {}", commit)?;
-            return Ok(1);
-        }
-    };
 
-    let main_branch_oid = repo.get_main_branch_oid()?;
-    let branch_oid_to_names = repo.get_branch_oid_to_names()?;
+    let references_snapshot = repo.get_references_snapshot()?;
     let conn = repo.get_db_conn()?;
-    let merge_base_db = MergeBaseDb::new(&conn)?;
     let event_log_db = EventLogDb::new(&conn)?;
     let event_replayer = EventReplayer::from_event_log_db(effects, &repo, &event_log_db)?;
     let event_cursor = event_replayer.make_default_cursor();
-    let graph = make_graph(
+    let mut dag = Dag::open_and_sync(
         effects,
         &repo,
-        &merge_base_db,
         &event_replayer,
         event_cursor,
-        &HeadOid(Some(source_oid)),
-        &MainBranchOid(main_branch_oid),
-        &BranchOids(branch_oid_to_names.keys().copied().collect()),
-        true,
+        &references_snapshot,
     )?;
 
+    let (source_oid, dest_oid) =
+        match resolve_commits(effects, &repo, &mut dag, vec![source, dest])? {
+            ResolveCommitsResult::Ok { commits } => match &commits.as_slice() {
+                [source_commit, dest_commit] => (source_commit.get_oid(), dest_commit.get_oid()),
+                _ => eyre::bail!("Unexpected number of returns values from resolve_commits"),
+            },
+            ResolveCommitsResult::CommitNotFound { commit } => {
+                writeln!(effects.get_output_stream(), "Commit not found: {}", commit)?;
+                return Ok(1);
+            }
+        };
+
     let source_oid = if should_resolve_base_commit {
-        let merge_base_oid =
-            merge_base_db.get_merge_base_oid(effects, &repo, source_oid, dest_oid)?;
-        resolve_base_commit(&graph, merge_base_oid, source_oid)
+        let merge_base_oid = dag.get_one_merge_base_oid(effects, &repo, source_oid, dest_oid)?;
+        resolve_base_commit(&dag, merge_base_oid, source_oid)?
     } else {
         source_oid
     };
 
+    let MoveOptions {
+        force_in_memory,
+        force_on_disk,
+        resolve_merge_conflicts,
+        dump_rebase_constraints,
+        dump_rebase_plan,
+    } = *move_options;
     let now = SystemTime::now();
     let event_tx_id = event_log_db.make_transaction_id(now, "move")?;
     let rebase_plan = {
-        let mut builder = RebasePlanBuilder::new(
-            &repo,
-            &graph,
-            &merge_base_db,
-            &MainBranchOid(main_branch_oid),
-        );
+        let mut builder = RebasePlanBuilder::new(&repo, &dag);
         builder.move_subtree(source_oid, dest_oid)?;
         builder.build(
             effects,
             &BuildRebasePlanOptions {
                 dump_rebase_constraints,
                 dump_rebase_plan,
-                detect_duplicate_commits_via_patch_id: false,
+                detect_duplicate_commits_via_patch_id: true,
             },
         )?
     };
     let result = match rebase_plan {
         Ok(None) => {
             writeln!(effects.get_output_stream(), "Nothing to do.")?;
-            0
+            return Ok(0);
         }
         Ok(Some(rebase_plan)) => {
             let options = ExecuteRebasePlanOptions {
@@ -159,13 +158,24 @@ pub fn r#move(
                 preserve_timestamps: get_restack_preserve_timestamps(&repo)?,
                 force_in_memory,
                 force_on_disk,
+                resolve_merge_conflicts,
             };
             execute_rebase_plan(effects, git_run_info, &repo, &rebase_plan, &options)?
         }
         Err(err) => {
             err.describe(effects, &repo)?;
-            1
+            return Ok(1);
         }
     };
-    Ok(result)
+
+    match result {
+        ExecuteRebasePlanResult::Succeeded => Ok(0),
+
+        ExecuteRebasePlanResult::DeclinedToMerge { merge_conflict } => {
+            merge_conflict.describe(effects, &repo)?;
+            Ok(1)
+        }
+
+        ExecuteRebasePlanResult::Failed { exit_code } => Ok(exit_code),
+    }
 }
