@@ -3,17 +3,19 @@
 //! This is inside `src` rather than `tests` since we use this code in some unit
 //! tests.
 
-use std::ffi::{OsStr, OsString};
+use std::collections::HashMap;
+use std::ffi::OsString;
+use std::fs;
 use std::io::Write;
 use std::ops::Deref;
-use std::path::PathBuf;
+use std::path::{Path, PathBuf};
 use std::process::{Command, Stdio};
 
 use crate::git::{GitRunInfo, GitVersion, NonZeroOid, Repo};
 use crate::util::get_sh;
 
-use color_eyre::Help;
 use eyre::{eyre, Context};
+use itertools::Itertools;
 use lazy_static::lazy_static;
 use regex::{Captures, Regex};
 use tempfile::TempDir;
@@ -57,7 +59,7 @@ impl Default for GitInitOptions {
 }
 
 /// Options for `Git::run_with_options`.
-#[derive(Debug)]
+#[derive(Debug, Default)]
 pub struct GitRunOptions {
     /// The timestamp of the command. Mostly useful for `git commit`. This should
     /// be a number like 0, 1, 2, 3...
@@ -68,16 +70,9 @@ pub struct GitRunOptions {
 
     /// The input to write to the child process's stdin.
     pub input: Option<String>,
-}
 
-impl Default for GitRunOptions {
-    fn default() -> Self {
-        GitRunOptions {
-            time: 0,
-            expected_exit_code: 0,
-            input: None,
-        }
-    }
+    /// Additional environment variables to start the process with.
+    pub env: HashMap<String, String>,
 }
 
 impl Git {
@@ -135,57 +130,47 @@ impl Git {
         Ok(output)
     }
 
+    fn get_git_exec_path(&self) -> PathBuf {
+        match std::env::var_os("GIT_EXEC_PATH") {
+            Some(git_exec_path) => git_exec_path.into(),
+            None => {
+                let git_path = self
+                    .path_to_git
+                    .parent()
+                    .expect("Unable to find git path parent");
+                git_path.to_path_buf()
+            }
+        }
+    }
+
     /// Get the `PATH` environment variable to use for testing.
     pub fn get_path_for_env(&self) -> OsString {
         let cargo_bin_path = assert_cmd::cargo::cargo_bin("git-branchless");
         let branchless_path = cargo_bin_path
             .parent()
             .expect("Unable to find git-branchless path parent");
-        let git_path = self
-            .path_to_git
-            .parent()
-            .expect("Unable to find git path parent");
         let bash = get_sh().expect("bash missing?");
         let bash_path = bash.parent().unwrap();
         std::env::join_paths(
             vec![
                 // For Git to be able to launch `git-branchless`.
-                branchless_path,
+                branchless_path.as_os_str(),
                 // For our hooks to be able to call back into `git`.
-                git_path,
+                self.get_git_exec_path().as_os_str(),
                 // For branchless to manually invoke bash when needed.
-                bash_path,
+                bash_path.as_os_str(),
             ]
-            .iter()
-            .map(|path| path.to_str().expect("Unable to decode path component")),
+            .into_iter(),
         )
         .expect("joining paths")
     }
 
-    /// Run a Git command.
-    #[instrument]
-    pub fn run_with_options<S: AsRef<str> + std::fmt::Debug>(
-        &self,
-        args: &[S],
-        options: &GitRunOptions,
-    ) -> eyre::Result<(String, String)> {
-        let GitRunOptions {
-            time,
-            expected_exit_code,
-            input,
-        } = options;
-
+    /// Get the environment variables needed to run git in the test environment.
+    pub fn get_base_env(&self, time: isize) -> Vec<(OsString, OsString)> {
         // Required for determinism, as these values will be baked into the commit
         // hash.
         let date: OsString = format!("{date} -{time:0>2}", date = DUMMY_DATE, time = time).into();
 
-        let args: Vec<&str> = {
-            let repo_path = self.repo_path.to_str().expect("Could not decode repo path");
-            let mut new_args: Vec<&str> = vec!["-C", repo_path];
-            new_args.extend(args.iter().map(|arg| arg.as_ref()));
-            new_args
-        };
-
         // Fake "editor" which accepts the default contents of any commit
         // messages. Usually, we can set this with `git commit -m`, but we have
         // no such option for things such as `git rebase`, which may call `git
@@ -194,17 +179,42 @@ impl Git {
         // ":" is understood by `git` to skip editing.
         let git_editor = OsString::from(":");
 
+        let git_exec_path = self.get_git_exec_path();
         let new_path = self.get_path_for_env();
-        let env: Vec<(&str, &OsStr)> = vec![
-            ("GIT_AUTHOR_DATE", &date),
-            ("GIT_COMMITTER_DATE", &date),
-            ("GIT_EDITOR", &git_editor),
-            ("PATH_TO_GIT", self.path_to_git.as_os_str()),
-            ("PATH", &new_path),
+        let envs = vec![
+            ("GIT_AUTHOR_DATE", date.clone()),
+            ("GIT_COMMITTER_DATE", date),
+            ("GIT_EDITOR", git_editor),
+            ("GIT_EXEC_PATH", git_exec_path.as_os_str().into()),
+            ("PATH_TO_GIT", self.path_to_git.as_os_str().into()),
+            ("PATH", new_path),
         ];
 
+        envs.into_iter()
+            .map(|(key, value)| (OsString::from(key), value))
+            .collect()
+    }
+
+    #[instrument]
+    fn run_with_options_inner(
+        &self,
+        args: &[&str],
+        options: &GitRunOptions,
+    ) -> eyre::Result<(String, String)> {
+        let GitRunOptions {
+            time,
+            expected_exit_code,
+            input,
+            env,
+        } = options;
+
         let mut command = Command::new(&self.path_to_git);
-        command.args(&args).env_clear().envs(env.iter().copied());
+        command
+            .current_dir(&self.repo_path)
+            .args(args)
+            .env_clear()
+            .envs(self.get_base_env(*time))
+            .envs(env.iter());
 
         let result = if let Some(input) = input {
             let mut child = command
@@ -263,6 +273,18 @@ impl Git {
         Ok((stdout, stderr))
     }
 
+    /// Run a Git command.
+    pub fn run_with_options<S: AsRef<str> + std::fmt::Debug>(
+        &self,
+        args: &[S],
+        options: &GitRunOptions,
+    ) -> eyre::Result<(String, String)> {
+        self.run_with_options_inner(
+            args.iter().map(|arg| arg.as_ref()).collect_vec().as_slice(),
+            options,
+        )
+    }
+
     /// Run a Git command.
     pub fn run<S: AsRef<str> + std::fmt::Debug>(
         &self,
@@ -284,7 +306,20 @@ impl Git {
         }
 
         // Non-deterministic metadata (depends on current time).
-        self.run(&["config", "branchless.commitMetadata.relativeTime", "false"])?;
+        self.run(&[
+            "config",
+            "branchless.commitDescriptors.relativeTime",
+            "false",
+        ])?;
+        self.run(&["config", "branchless.restack.preserveTimestamps", "true"])?;
+
+        // Disable warnings of the following form on Windows:
+        //
+        // ```
+        // warning: LF will be replaced by CRLF in initial.txt.
+        // The file will have its original line endings in your working directory
+        // ```
+        self.run(&["config", "core.autocrlf", "false"])?;
 
         if options.run_branchless_init {
             self.run(&["branchless", "init"])?;
@@ -299,13 +334,55 @@ impl Git {
         self.init_repo_with_options(&Default::default())
     }
 
+    /// Clone this repository into the `target` repository (which must not have
+    /// been initialized).
+    pub fn clone_repo_into(&self, target: &Git, additional_args: &[&str]) -> eyre::Result<()> {
+        let args = {
+            let mut args = vec![
+                "clone",
+                // For Windows in CI.
+                "-c",
+                "core.autocrlf=false",
+                self.repo_path.to_str().unwrap(),
+                target.repo_path.to_str().unwrap(),
+            ];
+            args.extend(additional_args.iter());
+            args
+        };
+
+        let (_stdout, _stderr) = self.run(args.as_slice())?;
+        Ok(())
+    }
+
     /// Write the provided contents to the provided file in the repository root.
     pub fn write_file(&self, name: &str, contents: &str) -> eyre::Result<()> {
+        let path = PathBuf::from(name);
+        if let Some(dir) = path.parent() {
+            std::fs::create_dir_all(self.repo_path.join(dir))?;
+        }
         let file_path = self.repo_path.join(format!("{}.txt", name));
         std::fs::write(&file_path, contents)?;
         Ok(())
     }
 
+    /// Delete the provided file in the repository root.
+    pub fn delete_file(&self, name: &str) -> eyre::Result<()> {
+        let file_path = self.repo_path.join(format!("{}.txt", name));
+        fs::remove_file(file_path)?;
+        Ok(())
+    }
+
+    /// Delete the provided file in the repository root.
+    pub fn set_file_permissions(
+        &self,
+        name: &str,
+        permissions: fs::Permissions,
+    ) -> eyre::Result<()> {
+        let file_path = self.repo_path.join(format!("{}.txt", name));
+        fs::set_permissions(file_path, permissions)?;
+        Ok(())
+    }
+
     /// Commit a file with default contents. The `time` argument is used to set
     /// the commit timestamp, which is factored into the commit hash.
     #[instrument]
@@ -395,6 +472,13 @@ impl Git {
         Ok(version >= GitVersion(2, 29, 0))
     }
 
+    /// The `log.excludeDecoration` configuration option was introduced in Git
+    /// v2.27.
+    pub fn supports_log_exclude_decoration(&self) -> eyre::Result<bool> {
+        let version = self.get_version()?;
+        Ok(version >= GitVersion(2, 27, 0))
+    }
+
     /// Resolve a file during a merge or rebase conflict with the provided
     /// contents.
     #[instrument]
@@ -414,17 +498,14 @@ impl Git {
 #[instrument]
 pub fn get_path_to_git() -> eyre::Result<PathBuf> {
     let path_to_git = std::env::var_os("PATH_TO_GIT")
-        .ok_or_else(|| eyre!("Could not get path to Git executable"))
-        .suggestion(
-            "No path to git set. Try running as: PATH_TO_GIT=$(which git) cargo test ...",
-        )?;
+        .ok_or_else(|| eyre!("No path to Git executable was set. Try running as: PATH_TO_GIT=$(which git) cargo test ..."))?;
     let path_to_git = PathBuf::from(&path_to_git);
     Ok(path_to_git)
 }
 
 /// Wrapper around a `Git` instance which cleans up the repository once dropped.
 pub struct GitWrapper {
-    _repo_dir: TempDir,
+    repo_dir: TempDir,
     git: Git,
 }
 
@@ -436,6 +517,39 @@ impl Deref for GitWrapper {
     }
 }
 
+/// From https://stackoverflow.com/a/65192210
+/// License: CC-BY-SA 4.0
+fn copy_dir_all(src: impl AsRef<Path>, dst: impl AsRef<Path>) -> std::io::Result<()> {
+    fs::create_dir_all(&dst)?;
+    for entry in fs::read_dir(src)? {
+        let entry = entry?;
+        let ty = entry.file_type()?;
+        if ty.is_dir() {
+            copy_dir_all(entry.path(), dst.as_ref().join(entry.file_name()))?;
+        } else {
+            fs::copy(entry.path(), dst.as_ref().join(entry.file_name()))?;
+        }
+    }
+    Ok(())
+}
+
+impl GitWrapper {
+    /// Make a copy of the repo on disk. This can be used to reuse testing
+    /// setup.  This is *not* the same as running `git clone`; it's used to save
+    /// initialization time as part of testing optimization.
+    ///
+    /// The copied repo will be deleted once the returned value has been dropped.
+    pub fn duplicate_repo(&self) -> eyre::Result<Self> {
+        let repo_dir = tempfile::tempdir()?;
+        copy_dir_all(&self.repo_dir, &repo_dir)?;
+        let git = Git {
+            repo_path: repo_dir.path().to_path_buf(),
+            path_to_git: self.git.path_to_git.clone(),
+        };
+        Ok(Self { repo_dir, git })
+    }
+}
+
 /// Create a temporary directory for testing and a `Git` instance to use with it.
 pub fn make_git() -> eyre::Result<GitWrapper> {
     let repo_dir = tempfile::tempdir()?;
@@ -446,8 +560,54 @@ pub fn make_git() -> eyre::Result<GitWrapper> {
         env: std::env::vars_os().collect(),
     };
     let git = Git::new(repo_dir.path().to_path_buf(), path_to_git);
-    Ok(GitWrapper {
-        _repo_dir: repo_dir,
-        git,
+    Ok(GitWrapper { repo_dir, git })
+}
+
+/// Represents a pair of directories that will be cleaned up after this value
+/// dropped. The two directories need to be `init`ed and `clone`ed by the
+/// caller, respectively.
+pub struct GitWrapperWithRemoteRepo {
+    /// Guard to clean up the containing temporary directory. Make sure to bind
+    /// this to a local variable not named `_`.
+    pub temp_dir: TempDir,
+
+    /// The wrapper around the original repository.
+    pub original_repo: Git,
+
+    /// The wrapper around the cloned repository.
+    pub cloned_repo: Git,
+}
+
+/// Create a `GitWrapperWithRemoteRepo`.
+pub fn make_git_with_remote_repo() -> eyre::Result<GitWrapperWithRemoteRepo> {
+    let path_to_git = get_path_to_git()?;
+    let temp_dir = tempfile::tempdir()?;
+    let git_run_info = GitRunInfo {
+        path_to_git,
+        working_directory: temp_dir.path().to_path_buf(),
+        env: Default::default(),
+    };
+    let original_repo_path = temp_dir.path().join("original");
+    std::fs::create_dir_all(&original_repo_path)?;
+    let original_repo = Git::new(
+        original_repo_path.clone(),
+        GitRunInfo {
+            working_directory: original_repo_path,
+            ..git_run_info.clone()
+        },
+    );
+    let cloned_repo_path = temp_dir.path().join("cloned");
+    let cloned_repo = Git::new(
+        cloned_repo_path.clone(),
+        GitRunInfo {
+            working_directory: cloned_repo_path,
+            ..git_run_info
+        },
+    );
+
+    Ok(GitWrapperWithRemoteRepo {
+        temp_dir,
+        original_repo,
+        cloned_repo,
     })
 }
